@@ -1,6 +1,6 @@
 # version set in setup.cfg
-#RELEASE = (1,1,1)
-#version_major, version_minor, version_patch = RELEASE
+RELEASE = (1,2,2)
+version_major, version_minor, version_patch = RELEASE
 
 from .reader import *
 from .utils import *
