@@ -1,13 +1,18 @@
 #!/usr/bin/env python3
 
-import sys, re, traceback
+import sys, re, traceback, os, time
+from concurrent.futures import ThreadPoolExecutor
 from ..abstract import AbstractReader
 from ..utils import *
 
 try:
-    import pymysql as mariadb
-except:
     import MySQLdb as mariadb
+except:
+    import pymysql as mariadb
+
+DEBUG = os.getenv('PYHDBPP_DEBUG') or '-v4' in sys.argv
+if DEBUG:
+    print('using %s as mariadb' % mariadb)
 
 # MariaDBReader
 
@@ -16,7 +21,7 @@ class MariaDBReader(AbstractReader):
     read-only API for hdb++ databases, based on PyTangoArchiving AbstractReader
     """
     
-    def __init__(self,config='',**kwargs):
+    def __init__(self, config='', **kwargs):
         """
         Arguments accepted by pymysql connections:
 
@@ -25,6 +30,7 @@ class MariaDBReader(AbstractReader):
         :param password: Password to use.
         :param database: Database to use, None to not use a particular one.
         :param port: MySQL port to use, default is usually OK. (default: 3306)
+        :param persistent=True: whether to use different mysql connection between queries (False required for multi-threading)
         :param bind_address: When the client has multiple network interfaces, specify
             the interface from which to connect to the host. Argument can be
             a hostname or an IP address.
@@ -63,7 +69,7 @@ class MariaDBReader(AbstractReader):
             The class needs an authenticate method taking an authentication packet as
             an argument.  For the dialog plugin, a prompt(echo, prompt) method can be used
             (if no authenticate method) for returning a string from the user. (experimental)
-        :param server_public_key: SHA256 authenticaiton plugin public key value. (default: None)
+        :param server_public_key: SHA256 authentication plugin public key value. (default: None)
         :param db: Alias for database. (for compatibility to MySQLdb)
         :param passwd: Alias for password. (for compatibility to MySQLdb)
         :param binary_prefix: Add _binary prefix on bytes and bytearray. (default: False)
@@ -74,15 +80,16 @@ class MariaDBReader(AbstractReader):
             
         self.config = config or {}
         self.config.update(kwargs)
-            
+        self.db = None
         self.database = self.config.get('database','hdbpp')
         self.user = self.config.get('user','')
         self.password = self.config.get('password','')
         self.port = int(self.config.get('port','3306'))
         self.host = self.config.get('host','localhost')
         self.decimate = self.config.get('decimate',None)
+        self.persistent = self.config.get('persistent',True)
         if self.decimate:
-            if re.match('[0-9]+',decimate):
+            if re.match('[0-9]+',self.decimate):
                 self.decimate = int(self.decimate)
             elif str(self.decimate).lower().strip() == 'false':
                 self.decimate = False
@@ -91,42 +98,76 @@ class MariaDBReader(AbstractReader):
 
         self.attributes = {}
         self.tables = {} #table creators
+        self.max_workers = 3
+        self.executor = ThreadPoolExecutor(max_workers=self.max_workers)
+        self.debug = kwargs.get('debug',False)
         
-        self._connect()
+        # Persistency is optional to allow multi-threading
+        if self.persistent:
+            self._trace('Creating persistent connection to DB')
+            self.db = self._connect()
         
     def __del__(self):
         if getattr(self,'_cursor',None):
             self._cursor.close()
         if getattr(self,'db',None):
             self.db.close()
+            
+    def _trace(self,*args,**kwargs):
+        if DEBUG or kwargs.get('error',False):
+            print(time.ctime(),self.host,self.database,*args)
         
     def _connect(self):
-        self.db = mariadb.connect(database=self.database,
-            user=self.user, password=self.password, port=self.port, 
-            host=self.host)
-        self._cursor = self.db.cursor()
-        return self._cursor
+        self._trace('{}.connect()'.format(self.database))
+        try:
+            db = mariadb.connect(database=self.database,
+                user=self.user, password=self.password, port=self.port, 
+                host=self.host)        
+        except: #Old Debian 9 API
+            db = mariadb.connect(db=self.database,
+                user=self.user, passwd=self.password, port=self.port, 
+                host=self.host)                    
+        return db
         
-    def _query(self,query,prune=False):
+    def _query(self, query, prune = False, renew = False):
         """
         query: SQL code
+        prune: if True, remove duplicated values (row[1:])
+        renew: create a new db connection, close on exit
         """
-        #print('%s.Query("%s")' % (self.database, query))
-        self._cursor.execute(query)
-        if prune:
-            r,l = [],True
-            while l:
-                try:
-                    l = self._cursor.fetchone()
-                    if l and (not r or l[1:] != r[-1][1:]):
-                        r.append(l)
-                except:
-                    print(r[-1:], l)
-                    traceback.print_exc()
-                    break
-            return r
-        else:
-            return self._cursor.fetchall()
+        renew = not self.db or renew
+        self._trace('{}.Query("{}",renew={})'.format(self.database, query, renew))
+        try:
+            db = self.db if not renew else self._connect()
+            _cursor = db.cursor()
+            _cursor.execute(query)
+            if prune:
+                r,l = [],True
+                while l:
+                    try:
+                        l = _cursor.fetchone()
+                        if l and (not r or l[1:] != r[-1][1:]):
+                            r.append(l)
+                    except:
+                        self._trace(r[-1:], l)
+                        traceback.print_exc()
+                        break
+                return r
+            else:
+                return _cursor.fetchall()
+        except Exception as e:
+            print('!'*80)
+            self._trace(query, e, error=True)
+            raise e
+        finally:
+            try:
+                _cursor.close()
+            except:
+                pass
+            try:
+                renew and db.close()
+            except:
+                pass
         
     def _describe_table(self,table):
         if not self.tables.get(table,None):
@@ -177,7 +218,7 @@ class MariaDBReader(AbstractReader):
         if pattern:
             return [a for a in self.attributes if attr_match(pattern,a)]
             
-        return self.attributes.keys()
+        return sorted(self.attributes.keys())
     
     def get_attribute_name(self,attribute):
         """
@@ -221,38 +262,193 @@ class MariaDBReader(AbstractReader):
         data = [(a,i,'att_'+t) for (a,i,t) in self._query(q)]               
         self.attributes.update((str(t[0]).lower(),t) for t in data)
         return data
+    
+    def get_attribute_frequency(self, attribute, n = 100, values = None):
+        values = values or self.get_last_attributes_values([attribute],n = n)
+        n = len(values)
+        if not n:
+            return 0
+        t0, t1 = values[0][0],values[-1][0]
+        if not abs(t0-t1):
+            return 0
+        return float(n)/abs(t0-t1)
 
-    def get_last_attributes_values(self, attributes, columns = '', n = 1):
+
+    def get_closer_attributes_values(self, attributes, time_bound=None, n=1, columns=[], cast = 'DOUBLE', **kwargs):
         """
-        Returns last values inserted in DB for a list of attributes
+        Returns the n closer values inserted around time in DB for a list of attributes.
+        If n is negative it will get the last inserted values, if positive it will get the next.
 
         arguments:
             attribute: fqdn for the attribute.
-            columns: requested columns separated by commas
+            columns: list of columns to query, default to data_time and value_r, quality, error.
+            time_bound: lower bound for data_time, if None should default to now()
+            n: number of samples, default to 1.
         returns:
-            {'att1':(epoch, r_value, w_value, quality, error_desc),
-             'att2':(epoch, r_value, w_value, quality, error_desc),
-             ...
-            }
+            [(epoch, r_value, w_value, quality, error_desc)]
         """
+        logger.debug('mariadb({}).get_closer_attributes_values({},{},{})'.format(self.database,attributes,time_bound,n))
         data = {}
-        columns = columns or 'data_time, value_r, quality, att_error_desc_id'
+        order = 'desc' if n<0 else ''
+        if not columns:
+            columns = 'data_time, value_r, quality, att_error_desc_id'
+        elif isSequence(columns):
+            columns = ','.join(columns)
+        columns = columns.replace(' ','')
+
+        if cast and 'data_time' in columns.split(',') and 'cast' not in columns.lower():
+            columns = columns.replace('data_time',
+                'CAST(UNIX_TIMESTAMP(data_time) AS %s)' % cast)
+        self._trace(columns)
         
-        for a in attributes:
+        for attr in attributes:
             try:
-                a,i,t = self.get_attribute_id_table(a)[0]
+                a,i,t = self.get_attribute_id_table(attr)[0]
                 tdesc = str(self._query('describe %s'%t))
                 tcol = ('int_time' if 'int_time' in tdesc else 'data_time')
                 cols = ','.join(c for c in columns.split(',') 
-                                if c.strip() in tdesc)
-                data[a] = self._query('select %s from %s where '
-                    'att_conf_id = %s order by %s desc limit %s'
-                    % (cols, t, i, tcol, n))
+                    if c.strip() in tdesc or 'cast' in c.lower())
+
+                where = 'att_conf_id = %s' % i
+                if time_bound:
+                    tb = time2type(time_bound,str)
+                    tb = self.str2mysqlsecs(tb) if tcol == 'int_time' else ("%s" % tb)
+                    if n<0:
+                        where += ' and %s <= "%s"' % (tcol, tb)
+                    else:
+                        where += ' and %s > "%s"' % (tcol, tb)
+
+                q = ('select %s from %s where %s order by %s %s limit %s'
+                        % (cols, t, where, tcol, 'desc' if n<0 else '', abs(n)))
+                self._trace(q)
+                data[attr] = self._query(q)
+                self._trace(attr,'done')
+
             except:
+                self._trace(traceback.format_exc())
                 raise Exception('AttributeNotFound: %s' % a) 
 
         return data
+
+    def get_mysqlsecsdiff(self,date=None):
+        """
+        Returns the value to be added to dates when querying int_time tables
+        """
+        if date is None: 
+            date = time2str()
+        if isinstance(date,(int,float)): 
+            date = time2str(date)
+        return self._query(
+            "select (TO_SECONDS('%s')-62167222800) - UNIX_TIMESTAMP('%s')" 
+            % (date,date))[0][0]
     
+    def str2mysqlsecs(self,date):
+        """ converts given date to int mysql seconds() value """
+        rt = str2time(date)
+        return int(rt+self.get_mysqlsecsdiff(date))
+    
+    def mysqlsecs2time(self,int_time,tref=0):
+        """ converts a mysql secons() value to epoch """
+        tref = tref or int_time
+        return int_time - self.get_mysqlsecsdiff(time2str(tref))    
+    
+    def _correct_dates(self, start_date, stop_date=None):
+        if stop_date is None:
+            stop_date = now()
+        start_date = time2type(start_date, str)
+        stop_date = time2type(stop_date, str)
+        return start_date, stop_date
+    
+    def _query_attribute_values(self, attribute, 
+                start_date, stop_date,
+                decimate=None, correlate=False, 
+                columns='', cast='DOUBLE', aggr=None,
+                **params):
+        """
+        Returns attribute values between start and stop dates.
+
+        arguments:
+            attribute: fqdn for the attribute.
+            start_date: datetime, beginning of the period to query.
+            stop_date: datetime, end of the period to query.
+                       if None, now() is used.
+            decimate: aggregation function to use in the form:
+                      {'timedelta0':(MIN, MAX, ...)
+                      , 'timedelta1':(AVG, COUNT, ...)
+                      , ...}
+                      if None, returns raw data.
+        returns:
+            [(epoch0, r_value, w_value, quality, error_desc),
+            (epoch1, r_value, w_value, quality, error_desc),
+            ... ]
+        """
+
+        a = attribute
+        columns = columns or 'data_time, value_r, quality, att_error_desc_id'
+        start_date,stop_date = self._correct_dates(start_date, stop_date)
+        renew = params.get('renew',params.get('asynch',False))
+
+        try:
+            a,i,t = self.get_attribute_id_table(a)[0]
+            self._trace('get_attribute_id_table({}) => {},{}'.format(a,i,t))
+            
+            int_time = 'int_time' in str(self._describe_table(t))
+            if isinstance(start_date, datetime.datetime):
+                start_date = date2str(start_date)
+                stop_date = date2str(stop_date)
+                
+            tb, te = self.str2mysqlsecs(start_date),self.str2mysqlsecs(stop_date)
+            b,e = (tb, te) if int_time else (
+                    "'%s'" % start_date, "'%s'" % stop_date)
+            
+            cols = ','.join(c for c in columns.split(',') 
+                            if c.strip() in str(self.tables[t]))
+            
+            if aggr and not "(value_r)" in cols:
+                if aggr.lower() == 'maxmin':
+                    cols = cols.replace('value_r', "max(value_r),min(value_r),")
+                else:
+                    cols = cols.replace('value_r', str(aggr)+'(value_r)')
+
+            if 'data_time,' in cols:
+                cols = cols.replace('data_time,',
+                    'CAST(UNIX_TIMESTAMP(data_time) AS %s),' % cast)
+
+            tcol = 'int_time' if int_time else 'data_time'
+            q = ('select %s from %s where '
+                'att_conf_id = %s and %s between %s and %s '
+                % (cols, t, i, tcol, b, e))
+            
+            if decimate is True:
+                # decimation set to limit buffer size to 1080p (plotting)
+                decimate = int((te-tb)/1080.)
+                
+            if decimate:
+                decimate = int(decimate)
+                q += ' and value_r is not NULL group by '
+                if int_time:
+                    q += '(%s DIV %d)' % ('int_time', decimate)
+                else:
+                    q += '(FLOOR(%s/%d))' % (
+                        'UNIX_TIMESTAMP(data_time)', decimate)
+                if 'array' in t:
+                    q += ',idx'
+                    
+            q += ' order by %s_time' % ('data','int')[int_time]
+            
+            self._trace(q)
+            # repeated integer values will be decimated here?
+            data = self._query(q, prune=decimate, renew = renew)
+            self._trace('obtained data: {}'.format(len(data),'L'))
+            if data:
+                self._trace(str(data[-1]))
+            return data
+            
+        except:
+            import traceback
+            traceback.print_exc()
+            return []
+            
     def get_attribute_values(self, attribute,
             start_date, stop_date=None,
             decimate=None,
@@ -277,10 +473,30 @@ class MariaDBReader(AbstractReader):
         """
         decimate = self.decimate if decimate is None else decimate
         attribute = self.get_attribute_name(attribute)
-        r = self.get_attributes_values([attribute], start_date, stop_date, 
-                                       decimate, False, **params)
-        #print(r.keys(),start_date,stop_date)
-        return r[attribute]
+        start_date, stop_date = self._correct_dates(start_date,stop_date)
+        fstart, fstop = str2time(start_date),str2time(stop_date)        
+
+        self._trace('get_attribute_values({},{},{},{},{})'
+            .format(attribute,start_date,stop_date,decimate,params))
+        if not params.get('split',False):
+            r = self.get_attributes_values([attribute], start_date, stop_date, 
+                                        decimate, **params)
+            #print(r.keys(),start_date,stop_date)
+            return r[attribute]
+        else:
+            delta = float(fstop-fstart)/self.max_workers
+            results,data = {},[]
+            for i in range(self.max_workers):
+                t0,t1 = fstart+i*delta,fstart+(i+1)*delta
+                self._trace(' subquery({},{})'.format(time2str(t0),time2str(t1)))
+                results[i] = self.executor.submit(
+                    self._query_attribute_values,attribute,t0,t1,
+                        decimate=decimate, correlate=False,
+                        params=params)
+            for i in range(self.max_workers):
+                self._trace(i,'done')
+                data.extend(results[i].result())
+            return data
 
     def get_attributes_values(self, attributes,
             start_date, stop_date=None,
@@ -316,70 +532,33 @@ class MariaDBReader(AbstractReader):
             ... ],
             'attr1':[(...),(...)]}
         """
-        data = {}
-        columns = columns or 'data_time, value_r, quality, att_error_desc_id'
-        if isinstance(start_date,(int,float)):
-            start_date = time2str(start_date) 
-        if stop_date is None:
-            stop_date = now()
-        if isinstance(stop_date,(int,float)):    
-            stop_date = time2str(stop_date)
-        
-        for a in attributes:
-            try:
-                a,i,t = self.get_attribute_id_table(a)[0]
-                #print('get_attribute_id_table({}) => {},{}'.format(a,i,t))
-                
-                int_time = 'int_time' in str(self._describe_table(t))
-                if isinstance(start_date, datetime.datetime):
-                    start_date = date2str(start_date)
-                    stop_date = date2str(stop_date)
-                    
-                tb, te = str2time(start_date),str2time(stop_date)
-                b,e = (tb, te) if int_time else (
-                        "'%s'" % start_date, "'%s'" % stop_date)
-                    
-                cols = ','.join(c for c in columns.split(',') 
-                                if c.strip() in str(self.tables[t]))
+        start_date, stop_date = self._correct_dates(start_date, stop_date)
+        self._trace('get_attributes_values',attributes,start_date,stop_date,correlate,columns,cast,params)
+        results,data = {},{}
+        self._trace(time.ctime(),time.time(),'preparing queries')
+        with ThreadPoolExecutor(max_workers=len(attributes)) as ex:
+            for a in attributes:
+                results[a] = ex.submit(
+                    self._query_attribute_values, a,
+                    start_date, stop_date, decimate=decimate,
+                    correlate=correlate, columns=columns, cast=cast,
+                    **params)
+            
+            dones = []
+            self._trace(time.ctime(),time.time(),'waiting results')
+            while True:
+                time.sleep(1e-02)
+                for a in attributes:
+                    if a not in dones and results[a].done():
+                        dones.append(a)
+                        self._trace(time.ctime(),time.time(),a+' done')
+                        data[a] = results[a].result()
+                        self._trace(a,len(data[a]))
 
-                if 'data_time,' in cols:
-                    cols = cols.replace('data_time,',
-                        'CAST(UNIX_TIMESTAMP(data_time) AS %s),' % cast)
+                if len(dones) == len(results):
+                    break
 
-                tcol = 'int_time' if int_time else 'data_time'
-                q = ('select %s from %s where '
-                    'att_conf_id = %s and %s between %s and %s '
-                    % (cols, t, i, tcol, b, e))
-                
-                if decimate is True:
-                    # decimation set to limit buffer size to 1080p (plotting)
-                    decimate = int((te-tb)/1080.)
-                    
-                if decimate:
-                    decimate = int(decimate)
-                    q += ' and value_r is not NULL group by '
-                    if int_time:
-                        q += '(%s DIV %d)' % ('int_time', decimate)
-                    else:
-                        q += '(FLOOR(%s/%d))' % (
-                            'UNIX_TIMESTAMP(data_time)', decimate)
-                    if 'array' in t:
-                        q += ',idx'
-                        
-                q += ' order by %s_time' % ('data','int')[int_time]
-                    
-                #print(q)
-                data[a] = self._query(q, prune=decimate)
-                #print('obtained data: {}'.format(len(data[a]),'L'))
-            except:
-                import traceback
-                traceback.print_exc()
-                data[a] = []
-
-        return data        
-        
-        return {'attr0': [(time.time(), 0., 0., 0, '')]
-                , 'attr1': [(time.time(), 0., 0., 0, '')]}
+        return data
     
     
 ##############################################################################
